@@ -28,6 +28,7 @@ import contextlib
 import datetime
 import errno
 import functools
+import gc
 import io
 import os
 import pathlib
@@ -46,6 +47,7 @@ from py7zr.compressor import SupportedMethods, get_methods_names
 from py7zr.exceptions import Bad7zFile, CrcError, DecompressionError, InternalError, UnsupportedCompressionMethodError
 from py7zr.helpers import ArchiveTimestamp, MemIO, NullIO, calculate_crc32, filetime_to_dt, readlink
 from py7zr.properties import DEFAULT_FILTERS, MAGIC_7Z, get_default_blocksize
+from py7zr.win32compat import is_windows_native_python, is_windows_unc_path
 
 if sys.platform.startswith("win"):
     import _winapi
@@ -130,17 +132,17 @@ class ArchiveFile:
     @property
     def archivable(self) -> bool:
         """File has a Windows `archive` flag."""
-        return self._test_attribute(stat.FILE_ATTRIBUTE_ARCHIVE)  # type: ignore  # noqa
+        return self._test_attribute(stat.FILE_ATTRIBUTE_ARCHIVE)  # noqa
 
     @property
     def is_directory(self) -> bool:
         """True if file is a directory, otherwise False."""
-        return self._test_attribute(stat.FILE_ATTRIBUTE_DIRECTORY)  # type: ignore  # noqa
+        return self._test_attribute(stat.FILE_ATTRIBUTE_DIRECTORY)  # noqa
 
     @property
     def readonly(self) -> bool:
         """True if file is readonly, otherwise False."""
-        return self._test_attribute(stat.FILE_ATTRIBUTE_READONLY)  # type: ignore  # noqa
+        return self._test_attribute(stat.FILE_ATTRIBUTE_READONLY)  # noqa
 
     def _get_unix_extension(self) -> Optional[int]:
         attributes = self._get_property("attributes")
@@ -161,14 +163,12 @@ class ArchiveFile:
         e = self._get_unix_extension()
         if e is not None:
             return stat.S_ISLNK(e)
-        return self._test_attribute(stat.FILE_ATTRIBUTE_REPARSE_POINT)  # type: ignore  # noqa
+        return self._test_attribute(stat.FILE_ATTRIBUTE_REPARSE_POINT)  # noqa
 
     @property
     def is_junction(self) -> bool:
         """True if file is a junction/reparse point on windows, otherwise False."""
-        return self._test_attribute(
-            stat.FILE_ATTRIBUTE_REPARSE_POINT | stat.FILE_ATTRIBUTE_DIRECTORY  # type: ignore  # noqa
-        )  # type: ignore  # noqa
+        return self._test_attribute(stat.FILE_ATTRIBUTE_REPARSE_POINT | stat.FILE_ATTRIBUTE_DIRECTORY)  # noqa  # noqa
 
     @property
     def is_socket(self) -> bool:
@@ -209,7 +209,7 @@ class ArchiveFileList(collections.abc.Iterable):
     """Iteratable container of ArchiveFile."""
 
     def __init__(self, offset: int = 0):
-        self.files_list = []  # type: List[dict]
+        self.files_list: List[dict] = []
         self.index = 0
         self.offset = offset
 
@@ -258,7 +258,7 @@ class ArchiveInfo:
         method_names: List[str],
         solid: bool,
         blocks: int,
-        uncompressed: int,
+        uncompressed: List[int],
     ):
         self.stat = stat
         self.filename = filename
@@ -309,6 +309,7 @@ class SevenZipFile(contextlib.AbstractContextManager):
     ) -> None:
         if mode not in ("r", "w", "x", "a"):
             raise ValueError("ZipFile requires mode 'r', 'w', 'x', or 'a'")
+        self.fp: BinaryIO
         self.mp = mp
         self.password_protected = password is not None
         if blocksize:
@@ -320,7 +321,7 @@ class SevenZipFile(contextlib.AbstractContextManager):
             self._filePassed: bool = False
             self.filename: str = file
             if mode == "r":
-                self.fp: BinaryIO = open(file, "rb")
+                self.fp = open(file, "rb")
             elif mode == "w":
                 self.fp = open(file, "w+b")
             elif mode == "x":
@@ -334,13 +335,13 @@ class SevenZipFile(contextlib.AbstractContextManager):
             self._filePassed = False
             self.filename = str(file)
             if mode == "r":
-                self.fp = file.open(mode="rb")  # type: ignore  # noqa   # typeshed issue: 2911
+                self.fp = file.open(mode="rb")  # noqa   # typeshed issue: 2911
             elif mode == "w":
-                self.fp = file.open(mode="w+b")  # type: ignore  # noqa
+                self.fp = file.open(mode="w+b")  # noqa
             elif mode == "x":
-                self.fp = file.open(mode="x+b")  # type: ignore  # noqa
+                self.fp = file.open(mode="x+b")  # noqa
             elif mode == "a":
-                self.fp = file.open(mode="r+b")  # type: ignore  # noqa
+                self.fp = file.open(mode="r+b")  # noqa
             else:
                 raise ValueError("File open error.")
             self.mode = mode
@@ -348,12 +349,12 @@ class SevenZipFile(contextlib.AbstractContextManager):
             self._filePassed = True
             self.fp = file
             self.filename = None
-            self.mode = mode  # type: ignore  #noqa
+            self.mode = mode  # noqa
         elif isinstance(file, io.IOBase):
             self._filePassed = True
             self.fp = file
             self.filename = getattr(file, "name", None)
-            self.mode = mode  # type: ignore  #noqa
+            self.mode = mode  # noqa
         else:
             raise TypeError("invalid file: {}".format(type(file)))
         self.encoded_header_mode = True
@@ -376,10 +377,10 @@ class SevenZipFile(contextlib.AbstractContextManager):
         except Exception as e:
             self._fpclose()
             raise e
-        self._dict = {}  # type: Dict[str, IO[Any]]
+        self._dict: Dict[str, IO[Any]] = {}
         self.dereference = dereference
-        self.reporterd = None  # type: Optional[Thread]
-        self.q = queue.Queue()  # type: queue.Queue[Any]
+        self.reporterd: Optional[Thread] = None
+        self.q: queue.Queue[Any] = queue.Queue()
 
     def __enter__(self):
         return self
@@ -397,7 +398,7 @@ class SevenZipFile(contextlib.AbstractContextManager):
         if not self._check_7zfile(self.fp):
             raise Bad7zFile("not a 7z file")
         self.sig_header = SignatureHeader.retrieve(self.fp)
-        self.afterheader = self.fp.tell()
+        self.afterheader: int = self.fp.tell()
         self.fp.seek(self.sig_header.nextheaderofs, os.SEEK_CUR)
         buffer = io.BytesIO(self.fp.read(self.sig_header.nextheadersize))
         if self.sig_header.nextheadercrc != calculate_crc32(buffer.getvalue()):
@@ -405,6 +406,7 @@ class SevenZipFile(contextlib.AbstractContextManager):
         header = Header.retrieve(self.fp, buffer, self.afterheader, password)
         if header is None:
             return
+        header._initilized = True
         self.header = header
         header.size += 32 + self.sig_header.nextheadersize
         buffer.close()
@@ -506,15 +508,15 @@ class SevenZipFile(contextlib.AbstractContextManager):
         return_dict: bool = False,
         callback: Optional[ExtractCallback] = None,
     ) -> Optional[Dict[str, IO[Any]]]:
-        if callback is not None and not isinstance(callback, ExtractCallback):
-            raise ValueError("Callback specified is not a subclass of py7zr.callbacks.ExtractCallback class")
-        elif callback is not None:
+        if callback is None:
+            pass
+        elif isinstance(callback, ExtractCallback):
             self.reporterd = Thread(target=self.reporter, args=(callback,), daemon=True)
             self.reporterd.start()
-        target_junction = []  # type: List[pathlib.Path]
-        target_sym = []  # type: List[pathlib.Path]
-        target_files = []  # type: List[Tuple[pathlib.Path, Dict[str, Any]]]
-        target_dirs = []  # type: List[pathlib.Path]
+        else:
+            raise ValueError("Callback specified is not an instance of subclass of py7zr.callbacks.ExtractCallback class")
+        target_files: List[Tuple[pathlib.Path, Dict[str, Any]]] = []
+        target_dirs: List[pathlib.Path] = []
         if path is not None:
             if isinstance(path, str):
                 path = pathlib.Path(path)
@@ -528,13 +530,9 @@ class SevenZipFile(contextlib.AbstractContextManager):
                     pass
                 else:
                     raise e
-        fnames = []  # type: List[str]  # check duplicated filename in one archive?
+        fnames: List[str] = []  # check duplicated filename in one archive?
         self.q.put(("pre", None, None))
         for f in self.files:
-            # TODO: sanity check
-            # check whether f.filename with invalid characters: '../'
-            if f.filename.startswith("../"):
-                raise Bad7zFile
             # When archive has a multiple files which have same name
             # To guarantee order of archive, multi-thread decompression becomes off.
             # Currently always overwrite by latter archives.
@@ -549,14 +547,34 @@ class SevenZipFile(contextlib.AbstractContextManager):
                         break
                     i += 1
             fnames.append(outname)
-            if path is not None:
-                outfilename = path.joinpath(outname)
+            # check f.filename has invalid directory traversals
+            if path is None:
+                # do following but is_relative_to introduced in py 3.9
+                # so I replaced it with relative_to. when condition is not satisfied, raise ValueError
+                # if not pathlib.Path(...).joinpath(outname.lstrip("./")).is_relative_to(...):
+                #    raise Bad7zFile
+                try:
+                    pathlib.Path(os.getcwd()).joinpath(outname.lstrip("./")).relative_to(os.getcwd())
+                except ValueError:
+                    raise Bad7zFile
+                outfilename = pathlib.Path(outname.lstrip("./"))
             else:
-                outfilename = pathlib.Path(outname)
-            if os.name == "nt":
-                if outfilename.is_absolute():
-                    # hack for microsoft windows path length limit < 255
-                    outfilename = pathlib.WindowsPath("\\\\?\\" + str(outfilename))
+                outfilename = path.joinpath(outname.lstrip("./"))
+                try:
+                    outfilename.relative_to(path)
+                except ValueError:
+                    raise Bad7zFile
+            # When python on Windows and not python on Cygwin,
+            # Add win32 file namespace to exceed microsoft windows
+            # path length limitation to 260 bytes
+            # ref.
+            # https://docs.microsoft.com/en-us/windows/win32/fileio/naming-a-file
+            # In editions of Windows before Windows 10 version 1607,
+            # the maximum length for a path is MAX_PATH, which is defined as
+            # 260 characters. In later versions of Windows, changing a registry key
+            # or select option when python installation is required to remove the limit.
+            if is_windows_native_python() and outfilename.is_absolute() and not is_windows_unc_path(outfilename):
+                outfilename = pathlib.WindowsPath("\\\\?\\" + str(outfilename))
             if targets is not None and f.filename not in targets:
                 self.worker.register_filelike(f.id, None)
                 continue
@@ -576,18 +594,8 @@ class SevenZipFile(contextlib.AbstractContextManager):
                 else:
                     pass
             elif f.is_socket:
-                pass
-            elif f.is_symlink:
-                target_sym.append(outfilename)
-                try:
-                    if outfilename.exists():
-                        outfilename.unlink()
-                except OSError as ose:
-                    if ose.errno not in [errno.ENOENT]:
-                        raise
-                self.worker.register_filelike(f.id, outfilename)
-            elif f.is_junction:
-                target_junction.append(outfilename)
+                pass  # TODO: implement me.
+            elif f.is_symlink or f.is_junction:
                 self.worker.register_filelike(f.id, outfilename)
             else:
                 self.worker.register_filelike(f.id, outfilename)
@@ -616,45 +624,29 @@ class SevenZipFile(contextlib.AbstractContextManager):
             )
 
         self.q.put(("post", None, None))
+        # early return when dict specified
         if return_dict:
             return self._dict
-        else:
-            # create symbolic links on target path as a working directory.
-            # if path is None, work on current working directory.
-            for t in target_sym:
-                sym_dst = t.resolve()
-                with sym_dst.open("rb") as b:
-                    sym_src = b.read().decode(encoding="utf-8")  # symlink target name stored in utf-8
-                sym_dst.unlink()  # unlink after close().
-                sym_dst.symlink_to(pathlib.Path(sym_src))
-            # create junction point only on windows platform
-            if sys.platform.startswith("win"):
-                for t in target_junction:
-                    junction_dst = t.resolve()
-                    with junction_dst.open("rb") as b:
-                        junction_target = pathlib.Path(b.read().decode(encoding="utf-8"))
-                        junction_dst.unlink()
-                        _winapi.CreateJunction(junction_target, str(junction_dst))  # type: ignore  # noqa
-            # set file properties
-            for outfilename, properties in target_files:
-                # mtime
-                lastmodified = None
-                try:
-                    lastmodified = ArchiveTimestamp(properties["lastwritetime"]).totimestamp()
-                except KeyError:
-                    pass
-                if lastmodified is not None:
-                    os.utime(str(outfilename), times=(lastmodified, lastmodified))
-                if os.name == "posix":
-                    st_mode = properties["posix_mode"]
-                    if st_mode is not None:
-                        outfilename.chmod(st_mode)
-                        continue
-                # fallback: only set readonly if specified
-                if properties["readonly"] and not properties["is_directory"]:
-                    ro_mask = 0o777 ^ (stat.S_IWRITE | stat.S_IWGRP | stat.S_IWOTH)
-                    outfilename.chmod(outfilename.stat().st_mode & ro_mask)
-            return None
+        # set file properties
+        for outfilename, properties in target_files:
+            # mtime
+            lastmodified = None
+            try:
+                lastmodified = ArchiveTimestamp(properties["lastwritetime"]).totimestamp()
+            except KeyError:
+                pass
+            if lastmodified is not None:
+                os.utime(str(outfilename), times=(lastmodified, lastmodified))
+            if os.name == "posix":
+                st_mode = properties["posix_mode"]
+                if st_mode is not None:
+                    outfilename.chmod(st_mode)
+                    continue
+            # fallback: only set readonly if specified
+            if properties["readonly"] and not properties["is_directory"]:
+                ro_mask = 0o777 ^ (stat.S_IWRITE | stat.S_IWGRP | stat.S_IWOTH)
+                outfilename.chmod(outfilename.stat().st_mode & ro_mask)
+        return None
 
     def _prepare_append(self, filters, password):
         if password is not None and filters is None:
@@ -663,15 +655,13 @@ class SevenZipFile(contextlib.AbstractContextManager):
             filters = DEFAULT_FILTERS.ARCHIVE_FILTER
         else:
             pass
-        folder = Folder()
-        folder.password = password
-        folder.prepare_coderinfo(filters)  # create compressor
-        self.header.main_streams.packinfo.enable_digests = False  # FIXME
-        self.header.main_streams.unpackinfo.folders.append(folder)
-        self.header.main_streams.unpackinfo.numfolders += 1
-        pos = self.afterheader + self.header.main_streams.packinfo.packpositions[-1]
+        self.header.filters = filters
+        self.header.password = password
+        if self.header.main_streams is not None:
+            pos = self.afterheader + self.header.main_streams.packinfo.packpositions[-1]
+        else:
+            pos = self.afterheader
         self.fp.seek(pos)
-        self.header.main_streams.substreamsinfo.num_unpackstreams_folders.append(0)
         self.worker = Worker(self.files, pos, self.header, self.mp)
 
     def _prepare_write(self, filters, password):
@@ -681,23 +671,18 @@ class SevenZipFile(contextlib.AbstractContextManager):
             filters = DEFAULT_FILTERS.ARCHIVE_FILTER
         else:
             pass
-        folder = Folder()
-        folder.password = password
-        folder.prepare_coderinfo(filters)
         self.files = ArchiveFileList()
         self.sig_header = SignatureHeader()
         self.sig_header._write_skelton(self.fp)
         self.afterheader = self.fp.tell()
-        self.header = Header.build_header([folder])
-        self.header.password = password
-        self.header.main_streams.packinfo.enable_digests = not self.password_protected  # FIXME
+        self.header = Header.build_header(filters, password)
         self.fp.seek(self.afterheader)
         self.worker = Worker(self.files, self.afterheader, self.header, self.mp)
-        self.worker.prepare_archive()
 
     def _write_flush(self):
-        folder = self.header.main_streams.unpackinfo.folders[-1]
-        self.worker.flush_archive(self.fp, folder)
+        if self.header._initialized:
+            folder = self.header.main_streams.unpackinfo.folders[-1]
+            self.worker.flush_archive(self.fp, folder)
         self._write_header()
 
     def _write_header(self):
@@ -781,16 +766,16 @@ class SevenZipFile(contextlib.AbstractContextManager):
             self.header_encryption = False
 
     @staticmethod
-    def _check_7zfile(fp: Union[BinaryIO, io.BufferedReader]) -> bool:
+    def _check_7zfile(fp: Union[BinaryIO, io.BufferedReader, io.IOBase]) -> bool:
         result = MAGIC_7Z == fp.read(len(MAGIC_7Z))[: len(MAGIC_7Z)]
         fp.seek(-len(MAGIC_7Z), 1)
         return result
 
-    def _get_method_names(self) -> str:
+    def _get_method_names(self) -> List[str]:
         try:
             return get_methods_names([folder.coders for folder in self.header.main_streams.unpackinfo.folders])
         except KeyError:
-            raise UnsupportedCompressionMethodError("Unknown method")
+            raise UnsupportedCompressionMethodError(self.header.main_streams.unpackinfo.folders, "Unknown method")
 
     def _read_digest(self, pos: int, size: int) -> int:
         self.fp.seek(pos)
@@ -810,71 +795,90 @@ class SevenZipFile(contextlib.AbstractContextManager):
 
     def _var_release(self):
         self._dict = None
-        self.files = None
-        self.header = None
-        self.worker = None
-        self.sig_header = None
+        self.worker.close()
+        del self.worker
+        del self.files
+        del self.header
+        del self.sig_header
+        gc.collect()
 
     @staticmethod
     def _make_file_info(target: pathlib.Path, arcname: Optional[str] = None, dereference=False) -> Dict[str, Any]:
-        f = {}  # type: Dict[str, Any]
+        f: Dict[str, Any] = {}
         f["origin"] = target
         if arcname is not None:
             f["filename"] = pathlib.Path(arcname).as_posix()
         else:
             f["filename"] = target.as_posix()
-        if os.name == "nt":
+        if sys.platform == "win32":
             fstat = target.lstat()
             if target.is_symlink():
                 if dereference:
                     fstat = target.stat()
                     if stat.S_ISDIR(fstat.st_mode):
                         f["emptystream"] = True
-                        f["attributes"] = fstat.st_file_attributes & FILE_ATTRIBUTE_WINDOWS_MASK  # type: ignore  # noqa
+                        f["attributes"] = fstat.st_file_attributes & FILE_ATTRIBUTE_WINDOWS_MASK  # noqa
                     else:
                         f["emptystream"] = False
-                        f["attributes"] = stat.FILE_ATTRIBUTE_ARCHIVE  # type: ignore  # noqa
+                        f["attributes"] = stat.FILE_ATTRIBUTE_ARCHIVE  # noqa
                         f["uncompressed"] = fstat.st_size
                 else:
                     f["emptystream"] = False
-                    f["attributes"] = fstat.st_file_attributes & FILE_ATTRIBUTE_WINDOWS_MASK  # type: ignore  # noqa
-                    # f['attributes'] |= stat.FILE_ATTRIBUTE_REPARSE_POINT  # type: ignore  # noqa
+                    f["attributes"] = fstat.st_file_attributes & FILE_ATTRIBUTE_WINDOWS_MASK  # noqa
+                    # TODO: handle junctions
+                    # f['attributes'] |= stat.FILE_ATTRIBUTE_REPARSE_POINT  # noqa
             elif target.is_dir():
                 f["emptystream"] = True
-                f["attributes"] = fstat.st_file_attributes & FILE_ATTRIBUTE_WINDOWS_MASK  # type: ignore  # noqa
+                f["attributes"] = fstat.st_file_attributes & FILE_ATTRIBUTE_WINDOWS_MASK  # noqa
             elif target.is_file():
                 f["emptystream"] = False
-                f["attributes"] = stat.FILE_ATTRIBUTE_ARCHIVE  # type: ignore  # noqa
+                f["attributes"] = stat.FILE_ATTRIBUTE_ARCHIVE  # noqa
                 f["uncompressed"] = fstat.st_size
-        else:
+        elif (
+            sys.platform == "darwin"
+            or sys.platform.startswith("linux")
+            or sys.platform.startswith("freebsd")
+            or sys.platform.startswith("netbsd")
+            or sys.platform.startswith("sunos")
+            or sys.platform == "aix"
+        ):
             fstat = target.lstat()
             if target.is_symlink():
                 if dereference:
                     fstat = target.stat()
                     if stat.S_ISDIR(fstat.st_mode):
                         f["emptystream"] = True
-                        f["attributes"] = stat.FILE_ATTRIBUTE_DIRECTORY  # type: ignore  # noqa
+                        f["attributes"] = stat.FILE_ATTRIBUTE_DIRECTORY  # noqa
                         f["attributes"] |= FILE_ATTRIBUTE_UNIX_EXTENSION | (stat.S_IFDIR << 16)
                         f["attributes"] |= stat.S_IMODE(fstat.st_mode) << 16
                     else:
                         f["emptystream"] = False
-                        f["attributes"] = stat.FILE_ATTRIBUTE_ARCHIVE  # type: ignore  # noqa
+                        f["attributes"] = stat.FILE_ATTRIBUTE_ARCHIVE  # noqa
                         f["attributes"] |= FILE_ATTRIBUTE_UNIX_EXTENSION | (stat.S_IMODE(fstat.st_mode) << 16)
                 else:
                     f["emptystream"] = False
-                    f["attributes"] = stat.FILE_ATTRIBUTE_ARCHIVE | stat.FILE_ATTRIBUTE_REPARSE_POINT  # type: ignore  # noqa
+                    f["attributes"] = stat.FILE_ATTRIBUTE_ARCHIVE | stat.FILE_ATTRIBUTE_REPARSE_POINT  # noqa
                     f["attributes"] |= FILE_ATTRIBUTE_UNIX_EXTENSION | (stat.S_IFLNK << 16)
                     f["attributes"] |= stat.S_IMODE(fstat.st_mode) << 16
             elif target.is_dir():
                 f["emptystream"] = True
-                f["attributes"] = stat.FILE_ATTRIBUTE_DIRECTORY  # type: ignore  # noqa
+                f["attributes"] = stat.FILE_ATTRIBUTE_DIRECTORY  # noqa
                 f["attributes"] |= FILE_ATTRIBUTE_UNIX_EXTENSION | (stat.S_IFDIR << 16)
                 f["attributes"] |= stat.S_IMODE(fstat.st_mode) << 16
             elif target.is_file():
                 f["emptystream"] = False
                 f["uncompressed"] = fstat.st_size
-                f["attributes"] = stat.FILE_ATTRIBUTE_ARCHIVE  # type: ignore  # noqa
+                f["attributes"] = stat.FILE_ATTRIBUTE_ARCHIVE  # noqa
                 f["attributes"] |= FILE_ATTRIBUTE_UNIX_EXTENSION | (stat.S_IMODE(fstat.st_mode) << 16)
+        else:
+            fstat = target.stat()
+            if target.is_dir():
+                f["emptystream"] = True
+                f["attributes"] = stat.FILE_ATTRIBUTE_DIRECTORY  # noqa
+            elif target.is_file():
+                f["emptystream"] = False
+                f["uncompressed"] = fstat.st_size
+                f["attributes"] = stat.FILE_ATTRIBUTE_ARCHIVE  # noqa
 
         f["creationtime"] = ArchiveTimestamp.from_datetime(fstat.st_ctime)
         f["lastwritetime"] = ArchiveTimestamp.from_datetime(fstat.st_mtime)
@@ -882,13 +886,13 @@ class SevenZipFile(contextlib.AbstractContextManager):
         return f
 
     def _make_file_info_from_name(self, bio, size: int, arcname: str) -> Dict[str, Any]:
-        f = {}  # type: Dict[str, Any]
+        f: Dict[str, Any] = {}
         f["origin"] = None
         f["data"] = bio
         f["filename"] = pathlib.Path(arcname).as_posix()
         f["uncompressed"] = size
         f["emptystream"] = size == 0
-        f["attributes"] = stat.FILE_ATTRIBUTE_ARCHIVE  # type: ignore  # noqa
+        f["attributes"] = stat.FILE_ATTRIBUTE_ARCHIVE  # noqa
         f["creationtime"] = ArchiveTimestamp.from_now()
         f["lastwritetime"] = ArchiveTimestamp.from_now()
         return f
@@ -925,8 +929,8 @@ class SevenZipFile(contextlib.AbstractContextManager):
 
     def list(self) -> List[FileInfo]:
         """Returns contents information"""
-        alist = []  # type: List[FileInfo]
-        lastmodified = None  # type: Optional[datetime.datetime]
+        alist: List[FileInfo] = []
+        lastmodified: Optional[datetime.datetime] = None
         for f in self.files:
             if f.lastwritetime is not None:
                 lastmodified = filetime_to_dt(f.lastwritetime)
@@ -963,7 +967,7 @@ class SevenZipFile(contextlib.AbstractContextManager):
     def reporter(self, callback: ExtractCallback):
         while True:
             try:
-                item = self.q.get(timeout=1)  # type: Optional[Tuple[str, str, str]]
+                item: Optional[Tuple[str, str, str]] = self.q.get(timeout=1)
             except queue.Empty:
                 pass
             else:
@@ -1002,6 +1006,7 @@ class SevenZipFile(contextlib.AbstractContextManager):
             path = file
         else:
             raise ValueError("Unsupported file type.")
+        self.header.initialize()
         file_info = self._make_file_info(path, arcname, self.dereference)
         self.header.files_info.files.append(file_info)
         self.header.files_info.emptyfiles.append(file_info["emptystream"])
@@ -1019,11 +1024,6 @@ class SevenZipFile(contextlib.AbstractContextManager):
         elif isinstance(bio, io.TextIOBase):
             # First check whether is it Text?
             raise ValueError("Unsupported file object type: please open file with Binary mode.")
-        elif hasattr(bio, "read") and hasattr(bio, "__sizeof__"):
-            # CPython's io.BufferedIOBase or io.BufferedReader has __sizeof__, but
-            # pypy3 don't have. So first check __sizeof__ and then goes to alternative.
-            # Also allowing objet type which has read() and length methods for duck typing
-            size = bio.__sizeof__()
         elif isinstance(bio, io.BufferedIOBase):
             # come here when subtype of io.BufferedIOBase that don't have __sizeof__ (eg. pypy)
             # alternative for `size = bio.__sizeof__()`
@@ -1034,12 +1034,19 @@ class SevenZipFile(contextlib.AbstractContextManager):
             size = last - current
         else:
             raise ValueError("Wrong argument passed for argument bio.")
-        file_info = self._make_file_info_from_name(bio, size, arcname)
-        self.header.files_info.files.append(file_info)
-        self.header.files_info.emptyfiles.append(file_info["emptystream"])
-        self.files.append(file_info)
-        folder = self.header.main_streams.unpackinfo.folders[-1]
-        self.worker.archive(self.fp, self.files, folder, deref=False)
+        if size > 0:
+            self.header.initialize()
+            file_info = self._make_file_info_from_name(bio, size, arcname)
+            self.header.files_info.files.append(file_info)
+            self.header.files_info.emptyfiles.append(file_info["emptystream"])
+            self.files.append(file_info)
+            folder = self.header.main_streams.unpackinfo.folders[-1]
+            self.worker.archive(self.fp, self.files, folder, deref=False)
+        else:
+            file_info = self._make_file_info_from_name(bio, size, arcname)
+            self.header.files_info.files.append(file_info)
+            self.header.files_info.emptyfiles.append(file_info["emptystream"])
+            self.files.append(file_info)
 
     def writestr(self, data: Union[str, bytes, bytearray, memoryview], arcname: str):
         if not isinstance(arcname, str):
@@ -1081,7 +1088,7 @@ class SevenZipFile(contextlib.AbstractContextManager):
     def test(self) -> Optional[bool]:
         self.fp.seek(self.afterheader)
         self.worker = Worker(self.files, self.afterheader, self.header, self.mp)
-        crcs = self.header.main_streams.packinfo.crcs  # type: Optional[List[int]]
+        crcs: Optional[List[int]] = self.header.main_streams.packinfo.crcs
         if crcs is None or len(crcs) == 0:
             return None
         packpos = self.afterheader + self.header.main_streams.packinfo.packpos
@@ -1118,13 +1125,15 @@ def is_7zfile(file: Union[BinaryIO, str, pathlib.Path]) -> bool:
     """
     result = False
     try:
-        if isinstance(file, io.IOBase) and hasattr(file, "read"):
-            result = SevenZipFile._check_7zfile(file)  # type: ignore  # noqa
+        if (isinstance(file, BinaryIO) or isinstance(file, io.BufferedReader) or isinstance(file, io.IOBase)) and hasattr(
+            file, "read"
+        ):
+            result = SevenZipFile._check_7zfile(file)
         elif isinstance(file, str):
             with open(file, "rb") as fp:
                 result = SevenZipFile._check_7zfile(fp)
         elif isinstance(file, pathlib.Path) or isinstance(file, pathlib.PosixPath) or isinstance(file, pathlib.WindowsPath):
-            with file.open(mode="rb") as fp:  # type: ignore  # noqa
+            with file.open(mode="rb") as fp:  # noqa
                 result = SevenZipFile._check_7zfile(fp)
         else:
             raise TypeError("invalid type: file should be str, pathlib.Path or BinaryIO, but {}".format(type(file)))
@@ -1152,14 +1161,14 @@ class Worker:
     """Extract worker class to invoke handler"""
 
     def __init__(self, files, src_start: int, header, mp=False) -> None:
-        self.target_filepath = {}  # type: Dict[int, Union[MemIO, pathlib.Path, None]]
+        self.target_filepath: Dict[int, Union[MemIO, pathlib.Path, None]] = {}
         self.files = files
         self.src_start = src_start
         self.header = header
         self.current_file_index = len(self.files)
         self.last_file_index = len(self.files)
         if mp:
-            self.concurrent = Process  # type: Union[Type[Thread], Type[Process]]
+            self.concurrent: Union[Type[Thread], Type[Process]] = Process
         else:
             self.concurrent = Thread
 
@@ -1196,10 +1205,12 @@ class Worker:
                             skip_notarget=skip_notarget,
                         )
                 else:
-                    filename = getattr(fp, "name", None)
+                    if getattr(fp, "name", None) is None:
+                        raise InternalError("Caught unknown variable status error")
+                    filename: str = getattr(fp, "name", "")  # do not become "" but it is for type check.
                     self.extract_single(open(filename, "rb"), empty_files, 0, 0, q)
                     concurrent_tasks = []
-                    exc_q = queue.Queue()  # type: queue.Queue
+                    exc_q: queue.Queue = queue.Queue()
                     for i in range(numfolders):
                         if skip_notarget:
                             if not any([self.target_filepath.get(f.id, None) for f in folders[i].files]):
@@ -1246,38 +1257,7 @@ class Worker:
             if isinstance(fp, str):
                 fp = open(fp, "rb")
             fp.seek(src_start)
-            just_check = []  # type: List[ArchiveFile]
-            for f in files:
-                if q is not None:
-                    q.put(
-                        (
-                            "s",
-                            str(f.filename),
-                            str(f.compressed) if f.compressed is not None else "0",
-                        )
-                    )
-                fileish = self.target_filepath.get(f.id, None)
-                if fileish is not None:
-                    # delayed execution of crc check.
-                    self._check(fp, just_check, src_end)
-                    just_check = []
-                    fileish.parent.mkdir(parents=True, exist_ok=True)
-                    with fileish.open(mode="wb") as ofp:
-                        if not f.emptystream:
-                            # extract to file
-                            crc32 = self.decompress(fp, f.folder, ofp, f.uncompressed, f.compressed, src_end)
-                            ofp.seek(0)
-                            if f.crc32 is not None and crc32 != f.crc32:
-                                raise CrcError(crc32, f.crc32, f.filename)
-                        else:
-                            pass  # just create empty file
-                elif not f.emptystream:
-                    just_check.append(f)
-                if q is not None:
-                    q.put(("e", str(f.filename), str(f.uncompressed)))
-            if not skip_notarget:
-                # delayed execution of crc check.
-                self._check(fp, just_check, src_end)
+            self._extract_single(fp, files, src_end, q, skip_notarget)
         except Exception as e:
             if exc_q is None:
                 raise e
@@ -1285,6 +1265,75 @@ class Worker:
                 exc_tuple = sys.exc_info()
                 exc_q.put(exc_tuple)
 
+    def _extract_single(
+        self,
+        fp: BinaryIO,
+        files,
+        src_end: int,
+        q: Optional[queue.Queue],
+        skip_notarget=True,
+    ) -> None:
+        """
+        Single thread extractor that takes file lists in single 7zip folder.
+        this may raise exception.
+        """
+        just_check: List[ArchiveFile] = []
+        for f in files:
+            if q is not None:
+                q.put(
+                    (
+                        "s",
+                        str(f.filename),
+                        str(f.compressed) if f.compressed is not None else "0",
+                    )
+                )
+            fileish = self.target_filepath.get(f.id, None)
+            if fileish is None:
+                if not f.emptystream:
+                    just_check.append(f)
+            else:
+                # delayed execution of crc check.
+                self._check(fp, just_check, src_end)
+                just_check = []
+                fileish.parent.mkdir(parents=True, exist_ok=True)
+                if not f.emptystream:
+                    if f.is_junction and not isinstance(fileish, MemIO) and sys.platform == "win32":
+                        with io.BytesIO() as ofp:
+                            self.decompress(fp, f.folder, ofp, f.uncompressed, f.compressed, src_end)
+                            dst: str = ofp.read().decode("utf-8")
+                            # fileish.unlink(missing_ok=True) > py3.7
+                            if fileish.exists():
+                                fileish.unlink()
+                            if sys.platform == "win32":  # hint for mypy
+                                _winapi.CreateJunction(str(fileish), dst)  # noqa
+                    elif f.is_symlink and not isinstance(fileish, MemIO):
+                        with io.BytesIO() as omfp:
+                            self.decompress(fp, f.folder, omfp, f.uncompressed, f.compressed, src_end)
+                            omfp.seek(0)
+                            sym_target = pathlib.Path(omfp.read().decode("utf-8"))
+                            # fileish.unlink(missing_ok=True) > py3.7
+                            if fileish.exists():
+                                fileish.unlink()
+                            fileish.symlink_to(sym_target)
+                    else:
+                        with fileish.open(mode="wb") as obfp:
+                            crc32 = self.decompress(fp, f.folder, obfp, f.uncompressed, f.compressed, src_end)
+                            obfp.seek(0)
+                            if f.crc32 is not None and crc32 != f.crc32:
+                                raise CrcError(crc32, f.crc32, f.filename)
+                else:
+                    # just create empty file
+                    if not isinstance(fileish, MemIO):
+                        fileish.touch()
+                    else:
+                        with fileish.open() as ofp:
+                            pass
+            if q is not None:
+                q.put(("e", str(f.filename), str(f.uncompressed)))
+        if not skip_notarget:
+            # delayed execution of crc check.
+            self._check(fp, just_check, src_end)
+
     def _check(self, fp, check_target, src_end):
         # delayed execution of crc check.
         for f in check_target:
@@ -1331,13 +1380,13 @@ class Worker:
 
     def _find_link_target(self, target):
         """Find the target member of a symlink or hardlink member in the archive."""
-        targetname = target.as_posix()  # type: str
+        targetname: str = target.as_posix()
         linkname = readlink(targetname)
         # Check windows full path symlinks
         if linkname.startswith("\\\\?\\"):
             linkname = linkname[4:]
         # normalize as posix style
-        linkname = pathlib.Path(linkname).as_posix()  # type: str
+        linkname: str = pathlib.Path(linkname).as_posix()
         member = None
         for j in range(len(self.files)):
             if linkname == self.files[j].origin.as_posix():
@@ -1364,8 +1413,8 @@ class Worker:
     def write(self, fp: BinaryIO, f, assym, folder):
         compressor = folder.get_compressor()
         if assym:
-            link_target = self._find_link_target(f.origin)  # type: str
-            tgt = link_target.encode("utf-8")  # type: bytes
+            link_target: str = self._find_link_target(f.origin)
+            tgt: bytes = link_target.encode("utf-8")
             fd = io.BytesIO(tgt)
             insize, foutsize, crc = compressor.compress(fd, fp)
             fd.close()
@@ -1379,14 +1428,6 @@ class Worker:
         insize, foutsize, crc = compressor.compress(f.data(), fp)
         return self._after_write(insize, foutsize, crc)
 
-    def prepare_archive(self):
-        self.header.main_streams.packinfo.numstreams = 0
-        self.header.main_streams.substreamsinfo.digests = []
-        self.header.main_streams.substreamsinfo.digestsdefined = []
-        self.header.main_streams.substreamsinfo.num_unpackstreams_folders = [0]
-        self.header.main_streams.packinfo.packsizes = []
-        self.header.main_streams.packinfo.crcs = []
-
     def flush_archive(self, fp, folder):
         compressor = folder.get_compressor()
         foutsize = compressor.flush(fp)
@@ -1420,3 +1461,8 @@ class Worker:
     def register_filelike(self, id: int, fileish: Union[MemIO, pathlib.Path, None]) -> None:
         """register file-ish to worker."""
         self.target_filepath[id] = fileish
+
+    def close(self):
+        del self.header
+        del self.files
+        del self.concurrent
